/*
 * lircrc.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <getopt.h>
#include <vdr/plugin.h>
#include <lirc/lirc_client.h>
#include <vdr/status.h>
#include <vdr/remote.h>
#ifndef LIRCRC_USER_FILE
#define LIRCRC_USER_FILE ".lircrc"
#endif

#define RECONNECTDELAY 3000 // ms

static const char *VERSION        = "0.0.2";
static const char *DESCRIPTION    = trNOOP("Adds LIRC support by using '.lircrc'");

class cThreadLircrc;

class cPluginLircrc : public cPlugin {
friend class cThreadLircrc;
private:
  // Add any member variables or functions you may need here.
  cThreadLircrc* m_pThreadLircrc;
  char*          m_pszLircrcProgArg;
  char*          m_pszLircrcFileArg;
  bool           m_bKillVDRsLircRemoteArg;
protected:
  struct lirc_config* m_pLircConfig;
public:
  cPluginLircrc(void);
  virtual ~cPluginLircrc();
  virtual const char *Version(void) { return VERSION; }
  virtual const char *Description(void) { return tr(DESCRIPTION); }
  virtual const char *CommandLineHelp(void);
  virtual bool ProcessArgs(int argc, char *argv[]);
  virtual bool Initialize(void);
  virtual bool Start(void);
};

class cThreadLircrc : protected cThread
{
friend class cPluginLircrc;
private:
  cPluginLircrc* m_pPluginLircrc;
  virtual void Action(void);
public:
  cThreadLircrc(cPluginLircrc* pPluginLircrc);
};

class cKillVDRsLircRemote : private cStatus
{
private:
  const char* m_pszPluginName;
public:
  cKillVDRsLircRemote(const char* pszPluginName);
private:
  virtual void OsdClear(void)
  {
    KillIt();
  }
  void KillIt(void);
};

cKillVDRsLircRemote::cKillVDRsLircRemote(const char* pszPluginName)
{
  m_pszPluginName = pszPluginName;
}

void cKillVDRsLircRemote::KillIt(void)
{
  cRemote* pRemote = Remotes.Last();
  while (pRemote)
  {
    if (strcmp(pRemote->Name(), "LIRC") == 0)
    {
      esyslog("%s: killing VDRs LIRC remote", m_pszPluginName);
      Remotes.Del(pRemote);
      break;
    }
    pRemote = Remotes.Prev(pRemote);
  }
  delete this;
}

cThreadLircrc::cThreadLircrc(cPluginLircrc* pPluginLircrc)
{
  m_pPluginLircrc = pPluginLircrc;
}

void cThreadLircrc::Action(void)
{
  while(true){

    int printerror = 1;

    esyslog("%s: not connected to lircd (%s), trying to connect every %.1f seconds", m_pPluginLircrc->Name(), m_pPluginLircrc->m_pszLircrcProgArg, float(RECONNECTDELAY/1000));
    while(!m_pPluginLircrc->m_pLircConfig){
      if(lirc_init(m_pPluginLircrc->m_pszLircrcProgArg, false) == -1){
        cCondWait::SleepMs(RECONNECTDELAY);
      } else {
        if(lirc_readconfig(m_pPluginLircrc->m_pszLircrcFileArg, &m_pPluginLircrc->m_pLircConfig, NULL)){
          if(printerror){
            printerror=0;
            esyslog("%s: error - can't read '%s', trying again every %.1f seconds", m_pPluginLircrc->Name(), m_pPluginLircrc->m_pszLircrcFileArg, float(RECONNECTDELAY/1000));
          }
          lirc_deinit();
          cCondWait::SleepMs(RECONNECTDELAY);
        }
      }
    }
    esyslog("%s: connected to lircd (%s)", m_pPluginLircrc->Name(), m_pPluginLircrc->m_pszLircrcProgArg);

    char *pszLircCode;
    char *pszLircConfig;
    int ret;
    int repeat;
    while(!lirc_nextcode(&pszLircCode))
    {
      if(pszLircCode == NULL)
        continue;

      if (sscanf(pszLircCode,"%*x %x", &repeat) != 1) repeat=0;

      while(!(ret = lirc_code2char(m_pPluginLircrc->m_pLircConfig, pszLircCode, &pszLircConfig)) && pszLircConfig)
      {
        char* pszNextKey = pszLircConfig;
        do
        {
          char* pszKey = pszNextKey;
          pszNextKey = strpbrk(pszLircConfig, " \t");
          if (pszNextKey)
            *(pszNextKey++) = '\0';
          eKeys key = cKey::FromString(pszKey);
          if (key != kNone)
            if (repeat > 0)
              cRemote::Put(eKeys(key | k_Repeat));
            else
              cRemote::Put(key);
        }
        while (pszNextKey);
      }
      free(pszLircCode);
      if(ret == -1) 
        break;
    }

    lirc_freeconfig(m_pPluginLircrc->m_pLircConfig);
    m_pPluginLircrc->m_pLircConfig=NULL;
    lirc_deinit();

  }
}

cPluginLircrc::cPluginLircrc(void)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
  m_pszLircrcProgArg = NULL;
  m_pszLircrcFileArg = NULL;
  m_bKillVDRsLircRemoteArg = true;
  m_pLircConfig = NULL;
  m_pThreadLircrc = NULL;
}

cPluginLircrc::~cPluginLircrc()
{
  // Clean up after yourself!
  if (m_pLircConfig)
  {
    delete m_pThreadLircrc;
    lirc_freeconfig(m_pLircConfig);
    lirc_deinit();
  }
}

const char *cPluginLircrc::CommandLineHelp(void)
{
  // Return a string that describes all known command line options.
  return "  -p <name>, --prog=<name>   name of 'prog' in .lircrc (default is 'VDR')\n"
         "  -k,        --kill          do not kill VDRs LIRC Remote (default is to kill it)\n"
         "  [<config_file>]            name and path of the config file (default is $HOME/.lircrc)\n";
}

bool cPluginLircrc::ProcessArgs(int argc, char *argv[])
{
  // Implement command line argument processing here if applicable.
  bool bMoreOptions = true;
  opterr = 0;
  do
  {
    static struct option aOptions[] =
    {
      { "prog", required_argument, 0, 'p' },
      { "kill", no_argument,       0, 'k' },
      { NULL }
    };
    switch(getopt_long(argc, argv, ":n:k", aOptions, NULL))
    {
    case 'p':
      m_pszLircrcProgArg = optarg;
      break;
    case 'k':
      m_bKillVDRsLircRemoteArg = false;
      break;
    case '?':
      esyslog("%s: error - unkown option %s [arg #%d] - ignored", Name(), argv[optind-1], optind-1);
      break;
    case ':':
      esyslog("%s: error - missing argument for option %s [arg #%d] - ignored", Name(), argv[optind-1], optind-1);
      break;
    default:
      bMoreOptions = false;
      break;
    }
  }
  while(bMoreOptions);
  if (optind < argc)
    m_pszLircrcFileArg = argv[optind++];
  if (optind < argc)
    esyslog("%s: error - extra arguments '%s'... [arg #%d...] - ignored", Name(), argv[optind], optind);
  return true;
}

bool cPluginLircrc::Initialize(void)
{
  // Initialize any background activities the plugin shall perform.
  if (!m_pszLircrcProgArg)
    m_pszLircrcProgArg = const_cast<char*>("VDR");

  if (!m_pszLircrcFileArg)
  {
    char* pszLircrcFile = NULL;
    char* pszHome = getenv("HOME");
    if (!pszHome) pszHome=const_cast<char*>("/");
    pszLircrcFile = (char*)malloc(strlen(pszHome)+1+strlen(LIRCRC_USER_FILE)+1);
    if (!pszLircrcFile){
      esyslog("%s: fatal error - out of memory", Name());
    } else {
      strcpy(pszLircrcFile, pszHome);
      if ((strlen(pszHome) > 0) && (pszLircrcFile[strlen(pszLircrcFile)-1] != '/'))
        strcat(pszLircrcFile,"/");
      strcat(pszLircrcFile, LIRCRC_USER_FILE);
      m_pszLircrcFileArg = pszLircrcFile;
    }
    if (pszLircrcFile)
      free(pszLircrcFile);
  }

  if (!m_pszLircrcFileArg) esyslog("%s: fatal error - no lircrc file given", Name());
  return true;
}

bool cPluginLircrc::Start(void)
{
  // Start any background activities the plugin shall perform.
  if (m_pszLircrcFileArg)
  {
    if (!(m_pThreadLircrc = new cThreadLircrc(this)))
      esyslog("%s: fatal error - out of memory", Name());
    else if (!m_pThreadLircrc->Start())
      esyslog("%s: fatal error - can't start thread", Name());
    else
    {
      esyslog("%s: background thread started for '%s' of '%s'", Name(), m_pszLircrcProgArg, m_pszLircrcFileArg);
      if (m_bKillVDRsLircRemoteArg)
        new cKillVDRsLircRemote(Name());
    }
  }
  return true;
}

VDRPLUGINCREATOR(cPluginLircrc); // Don't touch this!
